from functools import partial, wraps

from schemagic.core import validate_against_schema, validator
from schemagic.validators import date_string

ALWAYS = lambda: True
WHEN_DEBUGGING = lambda: __debug__
IDENTITY = lambda x: x

"""
NOTE EVERYTHING IN THIS FILE IS EXPERIMENTAL.  DO NOT EXPECT STABILITY OR USABILITY IN CURRENT FORM
"""

def validated(validation_predicate=None, coerce_data=True, input_schema=None, output_schema=None, fn=None):
    if fn is None:
        return partial(validated, validation_predicate, coerce_data, input_schema, output_schema)

    validation_predicate = validation_predicate or WHEN_DEBUGGING
    input_validator = validator(input_schema or IDENTITY, "input to function {0}".format(fn.__name__), validation_predicate=validation_predicate, coerce_data=coerce_data)
    output_validator = validator(output_schema or IDENTITY, "output from function {0}".format(fn.__name__), validation_predicate=validation_predicate, coerce_data=coerce_data)

    @wraps(fn)
    def _fn(*args, **kwargs):
        validated_args, validated_kwargs = validate_function_input(input_validator, args, kwargs)
        return output_validator(fn(*validated_args, **validated_kwargs))
    return _fn

def validate_function_input(input_validator, arg_list, kwarg_dict):
    if arg_list and kwarg_dict:
        validated_input = input_validator(arg_list + tuple([kwarg_dict]))
        args, kwargs = validated_input[:-1], validated_input[-1]
    elif kwarg_dict:
        validated_input = input_validator(kwarg_dict)
        args, kwargs = [], validated_input
    elif arg_list and len(arg_list) > 1:
        validated_input = input_validator(arg_list)
        args, kwargs = validated_input, {}
    elif arg_list and len(arg_list) is 1:
        validated_input = input_validator(arg_list[0])
        args, kwargs = [validated_input], {}
    else:
        args, kwargs = arg_list, kwarg_dict
    return args, kwargs


# from functools import wraps
#
# from flask.app import Flask
#
# from schemagic.core import validate_against_schema
# from schemagic.web import service_registry
#
# app = Flask(__name__)
# register_fibonnacci_services = service_registry(app)
#
#
# def memo(fn):
#     _cache = {}
#     @wraps(fn)
#     def _f(*args):
#         try:
#             return _cache[args]
#         except KeyError:
#             _cache[args] = result = fn(*args)
#             return result
#         except TypeError:
#             return fn(*args)
#     _f.cache = _cache
#     return _f
#
# @memo
# def fib(n):
#     if n == 0 or n == 1:
#         return 1
#     else:
#         return fib(n - 1) + fib(n - 2)
#
# register_fibonnacci_services(
#     dict(rule="/fibonacci",
#          input_schema={"n": int},
#          output_schema=int,
#          fn=fib))
#
# if __name__ == '__main__':
#     validate_against_schema({"n": int}, {"n": 6})
#     app.run(port=5000)
from schemagic.core import validate_against_schema
from schemagic.validators import null

validate_against_schema(null, "hello")



import json
from functools import partial, update_wrapper
import collections

from flask.globals import request
from flask.wrappers import Response

from schemagic.core import validate_against_schema, validator
from schemagic.utils import multiple_dispatch_fn

ALWAYS = lambda: True
WHEN_DEBUGGING = lambda: __debug__
IDENTITY = lambda x: x

dispatch_to_fn = multiple_dispatch_fn("dispatch_to_fn",{
    lambda fn, args: isinstance(args, basestring): lambda fn, arg_list: fn(arg_list),
    lambda fn, args: isinstance(args, collections.Sequence): lambda fn, arg_list: fn(*arg_list),
    lambda fn, args: isinstance(args, collections.MutableMapping): lambda fn, arg_list: fn(**arg_list)},
    default= lambda fn, arg_list: fn(arg_list)
)


def process_error(exception):
    if "input" in exception.message:
        return Response(
            response=exception,
            status=400)
    return Response(
        status=500,
        response=exception)


def webservice_fn(fn, input_validator, output_validator):
    try:
        return Response(
            response= reduce(lambda x, y: y(x),[
                json.loads,
                partial(validate_against_schema, input_validator),
                partial(dispatch_to_fn, fn),
                partial(validate_against_schema, output_validator),
                json.dumps
            ], request.data),
            status=200,
            mimetype="application/json"
        )
    except Exception as e:
        return process_error(e)


def service_route(service, validation_pred=None, coerce_data=True, rule=None, input_schema=None, output_schema=None, fn=None):
    if not rule:
        return update_wrapper(partial(service_route, service, validation_pred, coerce_data), service_route)
    if fn is None:
        return update_wrapper(partial(service_route, service, validation_pred, coerce_data, rule, input_schema, output_schema), service_route)

    validation_pred = validation_pred or WHEN_DEBUGGING
    input_validator = validator(input_schema or IDENTITY, "input to endpoint {0}".format(rule), validation_predicate=validation_pred, coerce_data=coerce_data)
    output_validator = validator(output_schema or IDENTITY, "output from endpoint {0}".format(rule), validation_predicate=validation_pred, coerce_data=coerce_data)

    service.add_url_rule(
        rule=rule,
        endpoint=fn.__name__ if hasattr(fn, "__name__") else rule,
        view_func=update_wrapper(lambda: webservice_fn(fn, input_validator, output_validator), fn),
        methods=['POST']
    )
    return fn


def service_registry(service, validation_pred=None, coerce_data=True, *service_definitions):
    if not service_definitions:
        return update_wrapper(partial(service_registry, service, validation_pred, coerce_data), service_registry)
    map(lambda definition: service_route(service, validation_pred, coerce_data, **definition), service_definitions)


import re
from functools import partial, update_wrapper

from schemagic.core import validate_against_schema
from schemagic.utils import merge


def predicate_validator(predicate, name=None, coercer=None, message=None, data=None):
    predicate.__name__ = name or predicate.__name__
    if data is None:
        return update_wrapper(partial(predicate_validator, predicate, name, coercer, message), predicate)
    data = coercer(data) if coercer else data
    message = (message or "data did not meet requirements of the predicate {0}".format(predicate.__name__)) + "\n value: {0}".format(data)
    if not predicate(data):
        raise ValueError(message)
    return data

formatted_string = lambda str_format, **kwargs: predicate_validator(
    lambda data: re.match(str_format, data),
    **merge(dict(name="formatted_string: {0}".format(str_format),
                 coercer=str,
                 message="string not of expected format: expected: {0}".format(format)),
            kwargs))

date_string = formatted_string(r'\d+\-\d+\-\d+')
datetime_string = formatted_string(r'\d+\-\d+\-\d+ \d+:\d+:\d+\.\d+')


null = predicate_validator(lambda val: val is None, name="null")
or_ = lambda *schemata: predicate_validator(
    lambda val: any(validate_against_schema(schema, val) for schema in schemata),
    name="any of schema's {0}".format(schemata),
)
enum = lambda *possible_vals: predicate_validator(
    lambda val: val in possible_vals,
    name="enumeration of allowable values: {0}".format(possible_vals),
)



import collections
from functools import partial

from schemagic.utils import merge_with, multiple_dispatch_fn

WHEN_DEBUGGING = lambda: __debug__

def validate_map_template(schema, value):
    key_schema, value_schema = schema.items()[0]
    validate_key_val_pair = lambda key, val: (validate_against_schema(key_schema, key), validate_against_schema(value_schema, val))
    return dict(map(validate_key_val_pair, value.keys(), value.values()))

def validate_keyed_mapping(schema, value):
    missing_keys = set(schema.keys()) - set(value.keys())
    if missing_keys:
        raise ValueError("Missing keys {0} for value {1}".format(missing_keys, value))
    return merge_with(validate_against_schema, schema, value)

def validate_sequence_template(schema, value):
    return map(schema[0], value)

def validate_strict_sequence(schema, value):
    if not len(schema) == len(value):
        raise ValueError("sequence has a different number of elements than its schema prescribes.  value: {0}, schema: {1}".format(value, schema))
    return map(lambda sub_schema, sub_value: validate_against_schema(sub_schema, sub_value), schema, value)

is_map_template = lambda schema: isinstance(schema, collections.MutableMapping) and len(schema.items()) is 1 and not isinstance(schema.keys()[0], str)
is_keyed_mapping = lambda schema: isinstance(schema, collections.MutableMapping) and not is_map_template(schema)
is_sequence_template = lambda schema: isinstance(schema, collections.Sequence) and len(schema) is 1
is_strict_sequence = lambda schema: isinstance(schema, collections.Sequence) and 1 < len(schema)

validate_against_schema = multiple_dispatch_fn( "validate_against_schema", {
    lambda schema, value: is_sequence_template(schema): validate_sequence_template,
    lambda schema, value: is_strict_sequence(schema): validate_strict_sequence,
    lambda schema, value: is_map_template(schema): validate_map_template,
    lambda schema, value: is_keyed_mapping(schema): validate_keyed_mapping},
    default=lambda schema, value: schema(value))

def validator(schema, subject_name_str, validation_predicate=None, coerce_data=False, data=None):
    if data is None:
        return partial(validator, schema, subject_name_str, validation_predicate, coerce_data)
    validation_predicate = validation_predicate or WHEN_DEBUGGING
    if not validation_predicate():
        return data
    try:
        coerced_and_validated_data = validate_against_schema(schema, data)
        return coerced_and_validated_data if coerce_data else data
    except Exception as e:
        message_details = {
            "subject": subject_name_str,
            "error": "{0}: {1}".format(e.__class__.__name__, e),
            "value": data,
            "schema": schema
        }
        raise ValueError("Bad value provided for {subject}. - error: {error} schema: {schema} value: {value}".format(**message_details))


import copy
import doctest
from contextlib import contextmanager
from functools import wraps, update_wrapper, partial

import operator

import itertools


def merge_with(fn, a, b):
    """ returns a new dictionary that is the merger of a and b.  applies fn to the values of colliding keys.

    >>> merge_with(operator.add, {"a": 1}, {"a": 1, "b": 1})
    {'a': 2, 'b': 1}
    """
    fresh_dict = copy.deepcopy(a)
    for k, v in b.items():
        if k in fresh_dict:
            fresh_dict[k] = fn(fresh_dict[k], v)
        else:
            fresh_dict[k] = v
    return fresh_dict

merge = partial(merge_with, lambda a, b: b)
merge.__name__ = "merge"

def multiple_dispatch_fn(name, dispatch_map, default=None):
    """
    creates a multiple dispatch function.

    returns a function whose implementation of the function is based on the arguments passed to it.  it decides
    what implementation to use by testing the arguments against a series of predicates to detect what situation
    is applicable.

    For example,
    >>> add_if_ints_multiply_if_floats = multiple_dispatch_fn(
    ... {lambda *nums: all(isinstance(num, int) for num in nums): operator.add,
    ...  lambda *nums: all(isinstance(num, float) for num in nums): operator.mul})
    >>> add_if_ints_multiply_if_floats(10, 10)
    20
    >>> add_if_ints_multiply_if_floats(10.0, 10.0)
    100.0

    You can also provide a default implementation to use if none of the predicates match.
    For example,
    >>> add_if_ints_else_return_unmodified = multiple_dispatch_fn(
    ... {lambda *nums: all(isinstance(num, int) for num in nums): operator.add},
    ...  default=lambda *items: items)
    >>> add_if_ints_else_return_unmodified(25, 25)
    50
    >>> add_if_ints_else_return_unmodified("hello", None)
    ('hello', None)

    :param dispatch_map: mapping {predicate_fn: implementation_fn}
    :param default: implementation fn to be used if none of the predicates are satisfied
    :return:
    """
    dispatch_map = dispatch_map or {}
    def _fn(*args, **kwargs):
        try:
            dispatch_fn = next(itertools.chain([value for key, value in dispatch_map.items() if key(*args, **kwargs)],
                                               [default] if default else []))
            return dispatch_fn(*args, **kwargs)
        except StopIteration:
            raise ValueError("No dispatch function found. args: {0}, kwargs: {1}".format(args, kwargs))
    _fn.__name__ = name
    return _fn

def separate_dict(initial_dict, *keys_to_remove):
    """returns 2 new dicts, one with some keys removed, and another with only those keys"""
    part1_keys, part2_keys = copy.copy(initial_dict), {}
    for key, val in part1_keys.items():
        if key in keys_to_remove:
            part2_keys[key] = val
            del part1_keys[key]
    return part1_keys, part2_keys

@contextmanager
def assert_raises(expected_error=None):
    try:
        yield
    except Exception as e:
        if expected_error:
            assert isinstance(e, expected_error), "Expected to raise type(s): {0}, raised {1} instead".format(expected_error, e.__class__.__name__)
        else:
            # No specific error raised, so we're good.
            pass
    else:
        raise AssertionError("No exception raised")

if __name__ == '__main__':
    doctest.testmod()


from setuptools import setup

NAME = 'schemagic'
PACKAGES = ['schemagic']
VERSION = '0.3.23'
LICENSE = "LGPL"
DESCRIPTION = 'Define the shape of your data with simple python data structures. Use those data descriptions to validate your application.'
AUTHOR = 'Tyler Tolton'
AUTHOR_EMAIL = 'tjtolton@gmail.com'
URL = 'https://github.com/TJTolton/schemagic'
KEYWORDS = ['schema', 'schemas', 'schemata',
          'validate', 'validation', 'validator'
          'json', 'REST', 'webservice', 'flask', 'POST'
          'agile']

if __name__ == "__main__":
    setup(
        name=NAME,
        description=DESCRIPTION,
        license=LICENSE,
        url=URL,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        keywords=KEYWORDS,
        packages=PACKAGES,
    )